# Copyright 2013, 2014, 2015 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import print_function, unicode_literals, absolute_import, division
import six

from nose.tools import istest

from reahl.tofu import Fixture, test, scenario, expected, NoException, vassert
from reahl.component.exceptions import ArgumentCheck, IncorrectArgumentError, arg_checks, IsInstance, IsSubclass, ArgumentCheckedCallable, NotYetAvailable
from reahl.component.decorators import deprecated

@istest
class ArgumentCheckTests(object):
    class ArgumentCheckScenarios(Fixture):
        def new_ModelObject(self):
            class ModelObject(object):
                @arg_checks(y=IsInstance(int), title=IsInstance(six.string_types))
                def do_something(self, x, y, title='a title', style=None):
                    pass
            return ModelObject
                        
        @scenario
        def incorrect_positional_arg(self):
            self.expected_exception = IsInstance
            self.args = (1, '2')
            self.call_args = self.args
            self.kwargs = dict(title='some title', style='style')
            self.callable = self.ModelObject().do_something

        @scenario
        def incorrect_kwarg(self):
            self.expected_exception = IsInstance
            self.args = (1, 2)
            self.call_args = self.args
            self.kwargs = dict(title=3, style='style')
            self.callable = self.ModelObject().do_something

        @scenario
        def incorrect_deduced_kwarg(self):
            self.expected_exception = IsInstance
            self.args = (1, 2, 2, 'style')
            self.call_args = self.args
            self.kwargs = {}
            self.callable = self.ModelObject().do_something

        @scenario
        def incorrect_kwarg_not_all_sent(self):
            self.expected_exception = IsInstance
            self.args = (1, 2)
            self.call_args = self.args
            self.kwargs = dict(title=None)
            self.callable = self.ModelObject().do_something

        @scenario
        def correct_args_defaulted_kwargs(self):
            self.expected_exception = NoException
            self.args = (1, 2)
            self.call_args = self.args
            self.kwargs = {}
            self.callable = self.ModelObject().do_something

        @scenario
        def correct_unbound_method_called(self):
            self.expected_exception = NoException
            self.args = (NotYetAvailable('self'), 1, 2)
            self.call_args = (self.ModelObject(), 1, 2)
            self.kwargs = {}
            self.callable = self.ModelObject.do_something

        @scenario
        def correct_function_called(self):
            @arg_checks(y=IsInstance(int), title=IsInstance(six.string_types))
            def do_something(x, y, title='a title', style=None):
                pass

            self.expected_exception = NoException
            self.args = (1, 2)
            self.call_args = (1, 2)
            self.kwargs = {}
            self.callable = do_something
            
        @scenario
        def checks_on_classmethod(self):
            self.expected_exception = IsInstance
            class ModelObject(object):
                @arg_checks(y=IsInstance(int), title=IsInstance(six.string_types))
                @classmethod
                def do_something(cls, x, y, title='a title', style=None):
                    pass
            self.args = (1, 'y')
            self.call_args = self.args
            self.kwargs = {}
            self.callable = ModelObject.do_something

        @scenario
        def wrong_args_sent(self):
            self.expected_exception = TypeError
            self.args = (1, 2, 3, 4, 5, 6, 7)
            self.call_args = self.args
            self.kwargs = {}
            self.callable = self.ModelObject().do_something


    @test(ArgumentCheckScenarios)
    def checking_arguments(self, fixture):
        """Methods can be augmented with argument checks. These checks are done when calling such a method,
           or before an actual call is made using ArgumentCheckedCallable.checkargs."""

        with expected(fixture.expected_exception):
            fixture.callable(*fixture.call_args, **fixture.kwargs)
        with expected(fixture.expected_exception):
            ArgumentCheckedCallable(fixture.callable).checkargs(*fixture.args, **fixture.kwargs)


        wrapped_exception = NoException
        if fixture.expected_exception is not NoException:
            wrapped_exception = IncorrectArgumentError

        with expected(wrapped_exception):
            ArgumentCheckedCallable(fixture.callable, explanation='some message').checkargs(*fixture.args, **fixture.kwargs)


    @test(Fixture)
    def stubbable_is_instance(self, fixture):
        """Classes can be marked with a flag to let them pass the IsInstance or IsSubclass checks even
           though they do not inherit from the specified class."""
           
        class A(object):
            pass
            
        class B(object):
            is_A = True
            
        with expected(NoException):
            vassert( IsInstance(A).is_valid(B()) )
        with expected(NoException):
            vassert( IsSubclass(A).is_valid(B) )
    
    @test(Fixture)
    def checking_before_args_available(self, fixture):
        """When checking arguments before a method call, the checks for some arguments may be ignored by passing NotYetAvailable
           instances for them. The name of NotYetAvailable arguments should match the name of the argument it replaces."""
        class ModelObject(object):
            @arg_checks(y=IsInstance(int), title=IsInstance(six.string_types))
            def do_something(self, x, y, title='a title', style=None):
                pass
                
        model_object = ModelObject()
        with expected(NoException):
            ArgumentCheckedCallable(model_object.do_something).checkargs('x', NotYetAvailable('y'), title='a title')
        with expected(IsInstance):
            ArgumentCheckedCallable(model_object.do_something).checkargs('x', NotYetAvailable('y'), title=123)

        with expected(IncorrectArgumentError):
            ArgumentCheckedCallable(model_object.do_something, explanation='an explanation').checkargs('x', NotYetAvailable('x'), title='a valid title')



@test(Fixture)
def argument_checks_with_deprecated_methods(fixture):
    """When used with @deprecated, argument checks still work."""
    @deprecated('this test class is deprecated', '1.2')
    class ADeprecatedClass(object):
        @arg_checks(y=IsInstance(int), title=IsInstance(six.string_types))
        def __init__(self, x, y, title='a title', style=None):
            pass

    with expected(IsInstance):
        ADeprecatedClass(1, 'y')
    with expected(IsInstance):
        ArgumentCheckedCallable(ADeprecatedClass).checkargs('x', 'y')
    with expected(IncorrectArgumentError):
        ArgumentCheckedCallable(ADeprecatedClass, explanation='an explanation').checkargs('x', NotYetAvailable('x'), title='a valid title')


    class ADeprecatedClass(object):
        @deprecated('this instance method is deprecated', '1.3')
        @arg_checks(y=IsInstance(int), title=IsInstance(six.string_types))
        def instance_method(self, x, y, title='a title', style=None):
            pass

        @deprecated('this class method is deprecated', '2.3')
        @arg_checks(y=IsInstance(int), title=IsInstance(six.string_types))
        @classmethod
        def class_method(cls, x, y, title='a title', style=None):
            pass

    with expected(IsInstance):
        ADeprecatedClass().instance_method('x', 'y')
    with expected(IsInstance):
        ADeprecatedClass.class_method('x', 'y')



# Copyright 2013, 2014 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Support for database schema migration."""

from __future__ import print_function, unicode_literals, absolute_import, division
from pkg_resources import parse_version
import logging
import warnings

from reahl.component.exceptions import ProgrammerError


class MigrationRun(object):
    def __init__(self, orm_control, eggs_in_order):
        self.orm_control = orm_control
        self.changes = MigrationSchedule('drop_fk', 'drop_pk', 'pre_alter', 'alter', 
                                         'create_pk', 'indexes', 'data', 'create_fk', 'cleanup')
        self.eggs_in_order = eggs_in_order

    def migrations_to_run_for(self, egg):
        return [migration(self.changes) 
                for migration in egg.compute_migrations(self.orm_control.schema_version_for(egg, default='0.0'))]

    def schedule_migrations(self):
        migrations_per_egg = [(egg, self.migrations_to_run_for(egg))
                              for egg in self.eggs_in_order]

        self.schedule_migration_changes(reversed(migrations_per_egg), 'upgrade')
        self.schedule_migration_changes(migrations_per_egg, 'upgrade_cleanup')
        self.schedule_migration_changes(migrations_per_egg, 'schedule_upgrades')

    def schedule_migration_changes(self, migrations_per_egg, method_name):
        for egg, migration_list in migrations_per_egg:
            current_schema_version = self.orm_control.schema_version_for(egg, default='0.0')
            message = 'Scheduling %s %s migrations for %s - from version %s to %s' % \
                          (len(migration_list), method_name, egg.name, 
                           current_schema_version, egg.version)
            logging.getLogger(__name__).info(message)
            for migration in migration_list:
                if (method_name in ['upgrade', 'upgrade_cleanup']) and hasattr(migration, method_name):
                    message = 'Please override Migration.schedule_upgrades() instead.'
                    warnings.warn('DEPRECATED: %s.%s. %s' % (migration.__class__.__name__, method_name, message), DeprecationWarning, stacklevel=-1)
                if hasattr(migration, method_name):
                    getattr(migration, method_name)()

    def execute_migrations(self):
        self.changes.execute_all()
        self.update_schema_versions()
        
    def update_schema_versions(self):
        for egg in self.eggs_in_order:
            if self.orm_control.schema_version_for(egg, default='0.0') != egg.version:
                logging.getLogger(__name__).info('Migrating %s - updating schema version to %s' % (egg.name, egg.version))
                self.orm_control.update_schema_version_for(egg)


class MigrationSchedule(object):
    def __init__(self, *phases):
        self.phases_in_order = phases
        self.phases = dict([(i, []) for i in phases])

    def schedule(self, phase, to_call, *args, **kwargs):
        try:
            self.phases[phase].append((to_call, args, kwargs))
        except KeyError as e:
            raise ProgrammerError('A phase with name<%s> does not exist.' % phase)

    def execute(self, phase):
        logging.getLogger(__name__).info('Executing schema change phase %s' % phase)
        for to_call, args, kwargs in self.phases[phase]:
            logging.getLogger(__name__).debug(' change: %s(%s, %s)' % (to_call.__name__, args, kwargs))
            to_call(*args, **kwargs)

    def execute_all(self):
        for phase in self.phases_in_order:
            self.execute(phase)


class Migration(object):
    """Represents one logical change that can be made to an existing database schema.
    
       You should extend this class to build your own domain-specific database schema migrations. Set the
       `version` class attribute to a string containing the version of your component for which this Migration
       should be run when upgrading from a previous version.
       
       Never use code imported from your component in a Migration, since Migration code is kept around in
       future versions of a component and may be run to migrate a schema with different versions of the code in your component.
    """
    version = None
    @classmethod
    def is_applicable(cls, current_schema_version, new_version):
        if not cls.version:
            raise ProgrammerError('Migration %s does not have a version set' % cls)
        return parse_version(cls.version) > parse_version(current_schema_version) and \
               parse_version(cls.version) <= parse_version(new_version)

    def __init__(self, changes):
        self.changes = changes

    def schedule(self, phase, to_call, *args, **kwargs):
        """Call this method to schedule a method call for execution later during the specified migration phase.

           Scheduled migrations are first collected from all components, then the calls scheduled for each defined
           phase are executed. Calls in one phase are executed in the order they were scheduled. Phases are executed
           in the following order:

           'drop_fk', 'drop_pk', 'pre_alter', 'alter', 'create_pk', 'indexes', 'data', 'create_fk', 'cleanup'

           :param phase: The name of the phase to schedule this call.
           :param to_call: The method or function to call.
           :param args: The positional arguments to be passed in the call.
           :param kwargs: The keyword arguments to be passed in the call.
        """
        self.changes.schedule(phase, to_call, *args, **kwargs)

    def schedule_upgrades(self):
        """Override this method in a subclass in order to supply custom logic for changing the database schema. This
           method will be called for each of the applicable Migrations listed for all components, in order of 
           dependency of components (the component deepest down in the dependency tree, first).

           **Added in 2.1.2**: Supply custom upgrade logic by calling `self.schedule()`.
        """
        warnings.warn('Ignoring %s.schedule_upgrades(): it does not override schedule_upgrades() (method name typo perhaps?)' % self.__class__.__name__ , 
                      UserWarning, stacklevel=-1)





# Copyright 2013, 2014, 2015 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


# memoized
# Adapted from the original:
# http://code.activestate.com/recipes/577452/ (r1)
# Original Copyright notice:
# The MIT License (MIT)
#
# Copyright (c) 2010 Daniel Miller
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


from __future__ import print_function, unicode_literals, absolute_import, division
import six

from functools import partial, wraps
import inspect
import warnings

import wrapt

@wrapt.decorator
def memoized(wrapped, instance, args, kwargs):
    cache_location = instance
    if instance is None:
        cache_location = wrapped

    try:
        cache = cache_location.__cache__
    except AttributeError:
        cache = cache_location.__cache__ = {}

    key = (wrapped, args[:], frozenset(kwargs.items()))
    try:
        res = cache[key]
    except KeyError:
        res = cache[key] = wrapped(*args, **kwargs)
    return res
        

def deprecated(message, version='n/a'):
    def catch_wrapped(f):
        def is_init_or_classmethod(member):
            if inspect.ismethod(member) and member.__self__ is f:
                return True
            return (inspect.ismethod(member) or inspect.isfunction(member)) and member.__name__ == '__init__'

        @wrapt.decorator
        def deprecated_wrapper(wrapped, instance, args, kwargs):
            deprecated_thing = wrapped.__self__ if is_init_or_classmethod(wrapped) else wrapped
            warnings.warn('DEPRECATED: %s. %s' % (deprecated_thing, message), DeprecationWarning, stacklevel=2)
            return wrapped(*args, **kwargs)

        if six.PY3 and f.__doc__:
            f.__doc__ = '%s\n.. deprecated:: %s\n   %s' % (f.__doc__, version, message)

        if inspect.isclass(f):
            for name, method in inspect.getmembers(f, predicate=is_init_or_classmethod):
                setattr(f, name, deprecated_wrapper(method))
            return f
        else:
            return deprecated_wrapper(f)

    return catch_wrapped





# Copyright 2013, 2014 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities to manipulate underlying databases - sometimes via an ORM tool."""

from __future__ import print_function, unicode_literals, absolute_import, division
import re
from contextlib import contextmanager
import logging

from reahl.component.eggs import ReahlEgg
from reahl.component.migration import MigrationSchedule, MigrationRun

class InvalidConnectionURIException(Exception):
    pass
    
    
class CouldNotFindDatabaseControlException(Exception):
    def __init__(self, url):
        super(CouldNotFindDatabaseControlException, self).__init__(url)
        self.url = url


class SystemControl(object):
    def __init__(self, config):
        self.config = config
        self.connection_uri = self.config.reahlsystem.connection_uri
        self.db_control = self.db_control_for_uri(self.connection_uri)
        self.orm_control = self.config.reahlsystem.orm_control
        self.orm_control.system_control = self

    def connect(self):
        self.orm_control.connect()
        if self.db_control.is_in_memory:
            self.create_db_tables()
        
    @contextmanager
    def auto_connected(self):
        if self.connected:
            yield
        
        self.connect()
        try:
            yield
        finally:
            self.disconnect()

    @property
    def connected(self):
        return self.orm_control.connected

    def disconnect(self):
        self.orm_control.disconnect()
        
    def managed_transaction(self):
        return self.orm_control.managed_transaction()

    def nested_transaction(self):
        return self.orm_control.nested_transaction()
                
    def finalise_session(self):
        self.orm_control.finalise_session()

    def set_transaction_and_connection(self, transaction):
        self.orm_control.set_transaction_and_connection(transaction)

    def commit(self):
        self.orm_control.commit()
        
    def rollback(self):
        self.orm_control.rollback()

    def db_control_for_uri(self, url):
        for i in self.config.reahlsystem.databasecontrols:
            if i.matches_uri(url):
                return i(url, self.config)
        raise CouldNotFindDatabaseControlException(url)

    def __getattr__(self, name):
        return getattr(self.db_control, name)

    def size_database(self):
        with self.orm_control.managed_transaction():
            return self.db_control.size_database(self.orm_control)

    def create_db_tables(self):
        eggs_in_order = ReahlEgg.get_all_relevant_interfaces(self.config.reahlsystem.root_egg)
        with self.orm_control.managed_transaction() as transaction:
            return self.orm_control.create_db_tables(transaction, eggs_in_order)

    def drop_db_tables(self):
        with self.orm_control.managed_transaction() as transaction:
            return self.orm_control.drop_db_tables(transaction)

    def initialise_database(self, yes=False):
        self.drop_database(yes=yes)
        self.create_database()
        self.connect()
        try:
            self.create_db_tables()
        finally:
            self.disconnect() 

    def migrate_db(self):
        eggs_in_order = ReahlEgg.get_all_relevant_interfaces(self.config.reahlsystem.root_egg)
        self.orm_control.migrate_db(eggs_in_order)
        return 0

    def diff_db(self):
        return self.orm_control.diff_db()

    def do_daily_maintenance(self):
        with self.orm_control.managed_transaction() as transaction:
            ReahlEgg.do_daily_maintenance_for_egg(self.config.reahlsystem.root_egg)


class DatabaseControl(object):
    def __init__(self, url, config):
        self.config = config
        self.connection_uri = url
        uri_parts = self.parse_connection_uri(url)
        self.user_name = uri_parts['user']
        self.password = uri_parts['password']
        self.host = uri_parts['host']
        port_string = uri_parts['port'] or '5432'
        self.port = int(port_string)
        self.database_name = uri_parts['database']

    def get_dbapi_connection_creator(self):
        return None

    @property    
    def is_in_memory(self):
        return False

    @classmethod
    def matches_uri(cls, url):
        try:
            cls.parse_connection_uri(url)
            return True
        except InvalidConnectionURIException:
            return False
    
    @classmethod
    def parse_connection_uri(cls, url):
        match = re.match(cls.uri_regex_string, url)
        if not match:
            raise InvalidConnectionURIException()
        return match.groupdict()


class NullDatabaseControl(DatabaseControl):
    uri_regex_string = r''
    @classmethod
    def parse_connection_uri(self, url):
        if not url:
            return {'user':'', 'password':'', 'host':'', 'port':'', 'database':''}
        raise InvalidConnectionURIException()

    def donothing(self, *args, **kwargs):
        pass

    def __getattr__(self, name):
        return self.donothing


class ORMControl(object):
    def migrate_db(self, eggs_in_order):
        with self.managed_transaction():
            migration_run = MigrationRun(self, eggs_in_order)
            migration_run.schedule_migrations()
            migration_run.execute_migrations()








# Copyright 2013, 2014 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The Reahl production commandline utility."""

from __future__ import print_function, unicode_literals, absolute_import, division

import pprint

from pkg_resources import DistributionNotFound

from reahl.component.dbutils import SystemControl
from reahl.component.shelltools import Command, ReahlCommandline
from reahl.component.context import ExecutionContext
from reahl.component.config import EntryPointClassList, Configuration, StoredConfiguration, MissingValue
from reahl.component.eggs import ReahlEgg




class ProdShellConfig(Configuration):
    commands = EntryPointClassList('reahl.component.prodcommands', description='The commands (classes) available to the production commandline shell')


class ProductionCommand(Command):
    """Superclass used for all production shell commands."""
    options = []
    usage_args = '<config_directory>'
    
    def execute(self, options, args):
        self.create_context(args[0])
        self.directory = args[0]

    def verify_commandline(self, options, args):
        if not len(args) >= 1:
            self.parser.error('No config directory given')

    def create_context(self, config_directory):
        try:
            self.context = ExecutionContext.for_config_directory(config_directory)
        except DistributionNotFound as ex:
            ex.args = ('%s (In development? Did you forget to do a "reahl setup -- develop -N"?)' % ex.args[0],)
            raise
        with self.context:
            self.context.system_control = SystemControl(self.context.config)

    @property
    def sys_control(self):
        return self.context.system_control
    
    @property
    def config(self):
        return self.context.config


class ListConfig(ProductionCommand):
    """Lists current configuration settings."""
    keyword = 'listconfig'
    options = ProductionCommand.options +\
               [('-v', '--values', dict(action='store_true', dest='print_values', help='prints the currently configured value')),
               ('-f', '--files', dict(action='store_true', dest='print_files', help='prints the filename where the setting should be defined')),
               ('-d', '--defaults', dict(action='store_true', dest='print_defaults', help='prints the default value')),
               ('-m', '--missing', dict(action='store_true', dest='print_missing_only', help='prints the missing values only')),
               ('-i', '--info', dict(action='store_true', dest='print_description', help='prints a description'))]

    def create_context(self, config_directory):
        self.context = ExecutionContext()

    def execute(self, options, args):
        super(ListConfig, self).execute(options, args)
        with self.context:
            print('Listing config for %s' % self.directory)
            config = StoredConfiguration(self.directory)
            config.configure(validate=False)
            for config_file, key, value, setting in config.list_all():
                to_print = '%-35s' % key
                if options.print_files:
                    to_print += '\t%s' % config_file
                if options.print_values:
                    to_print += '\t%s' % value
                if options.print_defaults:
                    if setting.defaulted:
                        message = setting.default
                        if setting.dangerous:
                            message += ' (DANGEROUS DEFAULT)'
                    elif setting.automatic:
                        message = 'AUTOMATIC'
                    else:
                        message = 'NO DEFAULT'
                    to_print += '\t%s' % message
                if options.print_description:
                    to_print += '\t%s' % setting.description

                if options.print_missing_only and not isinstance(value, MissingValue):
                    pass
                else:
                    print(to_print)


class CheckConfig(ProductionCommand):
    """Checks current configuration settings."""
    keyword = 'checkconfig'
    def execute(self, options, args):
        super(CheckConfig, self).execute(options, args)
        print('Checking config in %s' % self.directory)
        config = StoredConfiguration(self.directory)
        config.configure(validate=True)
        print('Config parses OK')


class CreateDBUser(ProductionCommand):
    """Creates the database user."""
    keyword = 'createdbuser'
    def execute(self, options, args):
        super(CreateDBUser, self).execute(options, args)
        return self.sys_control.create_db_user()


class DropDBUser(ProductionCommand):
    """Drops the database user."""
    keyword = 'dropdbuser'
    def execute(self, options, args):
        super(DropDBUser, self).execute(options, args)
        return self.sys_control.drop_db_user()


class DropDB(ProductionCommand):
    """Drops the database."""
    keyword = 'dropdb'
    options = [('-y', '--yes', dict(action='store_true', dest='yes',
                                      help='automatically answers yes on prompts'))]
    def execute(self, options, args):
        super(DropDB, self).execute(options, args)
        return self.sys_control.drop_database(yes=options.yes)


class CreateDB(ProductionCommand):
    """Creates the database."""
    keyword = 'createdb'
    def execute(self, options, args):
        super(CreateDB, self).execute(options, args)
        return self.sys_control.create_database()


class BackupDB(ProductionCommand):
    """Backs up the database."""
    keyword = 'backupdb'
    options = ProductionCommand.options +\
              [('-d', '--directory', dict(dest='directory', default='/tmp', help='the direcotry to back up to'))]
    def execute(self, options, args):
        super(BackupDB, self).execute(options, args)
        return self.sys_control.backup_database(options.directory)


class RestoreDB(ProductionCommand):
    """Restores up the database."""
    keyword = 'restoredb'
    options = ProductionCommand.options +\
              [('-f', '--filename', dict(dest='filename', default='/tmp/data.pgsql', help='the file to restore from'))]
    def execute(self, options, args):
        super(RestoreDB, self).execute(options, args)
        return self.sys_control.restore_database(options.filename)


class BackupAllDB(ProductionCommand):
    """Backs up all the databases on the host this project config points to."""
    keyword = 'backupall'
    options = ProductionCommand.options +\
              [('-d', '--directory', dict(dest='directory', default='/tmp', help='the direcotry to back up to'))]
    def execute(self, options, args):
        super(BackupAllDB, self).execute(options, args)
        return self.sys_control.backup_all_databases(options.directory)


class RestoreAllDB(ProductionCommand):
    """Restores all the databases on the host this project config points to."""
    keyword = 'restoreall'
    options = ProductionCommand.options +\
              [('-f', '--filename', dict(dest='filename', default='/tmp/data.sql', help='the file to restore from'))]
    def execute(self, options, args):
        super(RestoreAllDB, self).execute(options, args)
        return self.sys_control.restore_all_databases(options.filename)


class SizeDB(ProductionCommand):
    """Prints the current size of the database."""
    keyword = 'sizedb'
    def execute(self, options, args):
        super(SizeDB, self).execute(options, args)
        with self.context:
            with self.sys_control.auto_connected():
                print('Database size: %s' % self.sys_control.size_database())
        return 0


class CreateDBTables(ProductionCommand):
    """Creates all necessary tables in the database."""
    keyword = 'createdbtables'
    def execute(self, options, args):
        super(CreateDBTables, self).execute(options, args)
        with self.context:
            with self.sys_control.auto_connected():
                return self.sys_control.create_db_tables()


class DropDBTables(ProductionCommand):
    """Drops all necessary tables in the database."""
    keyword = 'dropdbtables'
    def execute(self, options, args):
        super(DropDBTables, self).execute(options, args)
        with self.context:
            with self.sys_control.auto_connected():
                return self.sys_control.drop_db_tables()


class MigrateDB(ProductionCommand):
    """Runs all necessary database migrations."""
    keyword = 'migratedb'
    def execute(self, options, args):
        super(MigrateDB, self).execute(options, args)
        with self.context:
            with self.sys_control.auto_connected():
                return self.sys_control.migrate_db()


class DiffDB(ProductionCommand):
    """Prints out a diff between the current database schema and what is expected by the current code."""
    keyword = 'diffdb'
    def execute(self, options, args):
        super(DiffDB, self).execute(options, args)
        with self.context:
            with self.sys_control.auto_connected():
                pprint.pprint(self.sys_control.diff_db(), indent=2, width=20)


class ListDependencies(ProductionCommand):
    """List all dependency eggs in dependency order."""
    keyword = 'listdeps'
    options = ProductionCommand.options +\
              [('-v', '--verbose', dict(action='store_true', dest='verbose', help='list direct dependencies too'))]
    def execute(self, options, args):
        super(ListDependencies, self).execute(options, args)
        with self.context:
            distributions = ReahlEgg.compute_ordered_dependent_distributions(self.config.reahlsystem.root_egg)
            for distribution in distributions:
                deps = ''
                if options.verbose:
                    deps = '[%s]' % (' | '.join([six.text_type(i) for i in distribution.requires()]))
                print('%s %s' % (distribution, deps))
        return 0


class RunJobs(ProductionCommand):
    """Runs all registered scripts."""
    keyword = 'runjobs'
    def execute(self, options, args):
        super(RunJobs, self).execute(options, args)
        with self.context:
            with self.sys_control.auto_connected():
                self.sys_control.do_daily_maintenance()
        return 0



class ProductionCommandline(ReahlCommandline):
    """The main class for invoking commands on projects in production environments."""
    def __init__(self, options):
        super(ProductionCommandline, self).__init__(options, ProdShellConfig())




# Copyright 2013, 2014 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import print_function, unicode_literals, absolute_import, division

import six
from contextlib import contextmanager
import warnings
import re

from reahl.tofu import Fixture, test, vassert, expected, NoException
from reahl.stubble import CallMonitor, EmptyStub


from reahl.component.dbutils import ORMControl
from reahl.component.eggs import ReahlEgg
from reahl.component.migration import Migration, MigrationSchedule, MigrationRun
from reahl.component.exceptions import ProgrammerError

class stubclass(object):
    def __init__(self, cls):
        pass
    def __call__(self, cls):
        warnings.warn('This needs to become stubble.stubclass, but stubble does not deal with this scenario - it needs to be fixed')
        return cls

@stubclass(ReahlEgg)
class ReahlEggStub(ReahlEgg):
    def __init__(self, name, version, migrations):
        super(ReahlEggStub, self).__init__(None)
        self._name = name
        self._version = version
        self.migrations = migrations
    @property
    def name(self):
        return self._name

    @property
    def version(self):
        return self._version

    @property
    def migrations_in_order(self):
        return self.migrations


@stubclass(ORMControl)
class ORMControlStub(ORMControl):
    created_schema_for = None

    def __init__(self):
        self.versions = {}

    @contextmanager
    def managed_transaction(self):
        yield

    def update_schema_version_for(self, egg):
        self.versions[egg.name] = egg.version

    def schema_version_for(self, egg, default=None):
        return self.versions[egg.name]

    def initialise_schema_version_for(self, egg):
        self.versions[egg.name] = egg.version

    def set_currently_installed_version_for(self, egg, version_number):
        self.versions[egg.name] = version_number



class MigrateFixture(Fixture):
    def new_orm_control(self):
        return ORMControlStub()


@test(MigrateFixture)
def how_migration_works(fixture):
    """Calls that will modify the database are scheduled in the schedule_upgrades() method of all
       the applicable Migrations for a single migration run. `shedule_upgrades()` is called on each
       migration in order of their versions. Once all calls are scheduled,
       they are executed as scheduled.
    """

    class SomeObject(object):
        calls_made = []
        def do_something(self, arg):
            self.calls_made.append(arg)
    some_object = SomeObject()
    
    class Migration1(Migration):
        version = '2.0'
        def schedule_upgrades(self):
            self.schedule('drop_fk', some_object.do_something, 'drop_fk_1')
            self.schedule('data', some_object.do_something, 'data_1')
            self.schedule('drop_fk', some_object.do_something, 'drop_fk_2')

    class Migration2(Migration):
        version = '3.0'
        def schedule_upgrades(self):
            self.schedule('drop_fk', some_object.do_something, 'drop_fk_3')

    egg = ReahlEggStub('my_egg', '4.0', [Migration1, Migration2])
    fixture.orm_control.set_currently_installed_version_for(egg, '1.0')

    fixture.orm_control.migrate_db([egg])

    expected_order = ['drop_fk_1', 'drop_fk_2', 'drop_fk_3', 'data_1']
    vassert( some_object.calls_made == expected_order )


@test(MigrateFixture)
def schedule_executes_in_order(fixture):
    """A MigrationSchedule is used internally to schedule calls in different phases. The calls 
       scheduled in each phase are executed in the order the phases have been set up on the MigrationSchedule.
       Within a phase, the calls are executed in the order they were registered in that phase.
    """
    
    schedule_names = ['a', 'b', 'c']
    migration_schedule = MigrationSchedule(*schedule_names)

    class SomeObject(object):
        def do_something(self, arg):
            pass
    some_object = SomeObject()

    #schedule calls not in registered order
    with CallMonitor(some_object.do_something) as monitor:
        migration_schedule.schedule('c', some_object.do_something, 'c1')
        migration_schedule.schedule('a', some_object.do_something, 'a1')
        migration_schedule.schedule('b', some_object.do_something, 'b')
        migration_schedule.schedule('a', some_object.do_something, 'a2')
        migration_schedule.schedule('c', some_object.do_something, 'c2')

    migration_schedule.execute_all()

    actual_order = [call.args[0] for call in monitor.calls]
    expected_order = ['a1', 'a2', 'b', 'c1', 'c2']
    vassert( actual_order == expected_order )


@test(MigrateFixture)
def schedule_executes_phases_with_parameters(fixture):
    """When a MigrationSchedule executes the calls that were scheduled from a Migration, 
       the methods are actually called, and passed the correct arguments."""

    class SomeObject(object):
        def please_call_me(self, arg, kwarg=None):
            pass
    some_object = SomeObject()
    
    migration_schedule = MigrationSchedule('phase_name')
    migration = Migration(migration_schedule)

    with CallMonitor(some_object.please_call_me) as monitor:
        migration.schedule('phase_name', some_object.please_call_me, 'myarg', kwarg='mykwarg')

    migration_schedule.execute_all()

    vassert( monitor.calls[0].args == ('myarg',) )
    vassert( monitor.calls[0].kwargs == dict(kwarg='mykwarg') )


@test(MigrateFixture)
def invalid_schedule_name_raises(fixture):
    """A useful error is raised when an attempt is made to schedule a call in a phase that is not defined."""
    
    valid_schedule_names = ['a', 'b']
    migration_schedule = MigrationSchedule(*valid_schedule_names)

    def check_exception(ex):
        vassert( six.text_type(ex) == 'A phase with name<wrong_name> does not exist.' )

    with expected(ProgrammerError, test=check_exception):
        migration_schedule.schedule('wrong_name', None)


@test(MigrateFixture)
def version_dictates_execution_of_migration_(fixture):
    """Each Migration should have a class attribute `version` that states which version of the component
       it upgrades the database schema to. Only the Migrations with versions greater than the current 
       schema version are included in a MigrationRun for a given egg.
    """
    
    class PreviousVersionMigration(Migration):
        version = '1.0'
    class MatchingCurrentVersionMigration(Migration):
        version = '2.0'
    class NewerVersionMigration(Migration):
        version = '3.0'
    class EvenNewerVersionMigration(Migration):
        version = '4.0'

    egg = ReahlEggStub('my_egg', '4.0', [PreviousVersionMigration, MatchingCurrentVersionMigration, 
                                         NewerVersionMigration, EvenNewerVersionMigration])
    fixture.orm_control.set_currently_installed_version_for(egg, '2.0')

    migration_run = MigrationRun(fixture.orm_control, [egg])
    migrations_to_run = migration_run.migrations_to_run_for(egg)
    classes_to_run = [m.__class__ for m in migrations_to_run]
    vassert( classes_to_run == [NewerVersionMigration, EvenNewerVersionMigration] )


@test(MigrateFixture)
def version_of_migration_not_set_error(fixture):
    """If the version to which a Migration is applicable is not set, an error is raised."""
    class TestMigration(Migration):
        pass

    egg = ReahlEggStub('my_egg', '1.0', [TestMigration])
    fixture.orm_control.set_currently_installed_version_for(egg, '0.0')

    def check_exception(ex):
        vassert( re.match('Migration <class \'reahl\.component_dev\.migrationtests\..*TestMigration\'> does not have a version set', six.text_type(ex) ))

    with expected(ProgrammerError, test=check_exception):
        fixture.orm_control.migrate_db([egg])


@test(MigrateFixture)
def missing_schedule_upgrades_warns(fixture):
    """If a programmer does not override schedule_upgrades, a warning is raised."""
    class TestMigration(Migration):
        pass

    with warnings.catch_warnings(record=True) as raised_warnings:
         warnings.simplefilter("always")
  
         TestMigration(EmptyStub()).schedule_upgrades()
 
    [warning] = raised_warnings
    expected_message = 'Ignoring TestMigration.schedule_upgrades(): it does not override schedule_upgrades() (method name typo perhaps?)'
    vassert( six.text_type(warning.message) == expected_message  )


@test(MigrateFixture)
def available_migration_phases(fixture):
    """These are the phases, and order of the phases in a MigrationRun."""

    migration_run = MigrationRun(fixture.orm_control, [])
    
    expected_order = ('drop_fk', 'drop_pk', 'pre_alter', 'alter', 'create_pk', 'indexes', 'data', 'create_fk', 'cleanup')
    vassert( migration_run.changes.phases_in_order == expected_order  )


@test(MigrateFixture)
def schema_version_housekeeping(fixture):
    """The database keeps track of the schema for each installed component. After a migration run
       the currently installed versions are updated.
    """
    
    egg = ReahlEggStub('my_egg', '2.0', [])
    fixture.orm_control.set_currently_installed_version_for(egg, '1.0')
    migration_run = MigrationRun(fixture.orm_control, [egg])
    migration_run.execute_migrations()
    vassert( fixture.orm_control.schema_version_for(egg) == '2.0' )






# Copyright 2014 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import print_function, unicode_literals, absolute_import, division

import six


def ascii_as_bytes_or_str(unicode_str):
    if six.PY2:
        return unicode_str.encode('ascii')
    else:
        return unicode_str


def _html_escape_function():
    if six.PY2:
        import cgi
        def html_escape(s, quote=True):
            return cgi.escape(s, quote=quote)
        return html_escape
    else:
        import html
        return html.escape


html_escape = _html_escape_function()



from setuptools import setup, Command
class InstallTestDependencies(Command):
    user_options = []
    def run(self):
        from setuptools.command import easy_install
        easy_install.main(self.distribution.tests_require)

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

setup(
    name='reahl-component',
    version='3.1.1',
    description='The component framework of Reahl.',
    long_description="Reahl is a web application framework that allows a Python programmer to work in terms of useful abstractions - using a single programming language.\n\nReahl-component is the component that contains Reahl's component framework.\n\nSee http://www.reahl.org/docs/3.1/tutorial/gettingstarted.d.html for installation instructions. ",
    url='http://www.reahl.org',
    maintainer='Iwan Vosloo',
    maintainer_email='iwan@reahl.org',
    packages=['reahl', 'reahl.component', 'reahl.messages', 'reahl.component_dev'],
    py_modules=['setup'],
    include_package_data=False,
    package_data={'': ['*/LC_MESSAGES/*.mo']},
    namespace_packages=['reahl', 'reahl.messages'],
    install_requires=['six', 'Babel>=2.1,<2.1.999', 'python-dateutil>=2.2,<2.2.999', 'wrapt>=1.10.2,<1.10.999'],
    setup_requires=['reahl-bzrsupport>=3.1,<3.2'],
    tests_require=['nose', 'reahl-tofu>=3.1,<3.2', 'reahl-stubble>=3.1,<3.2'],
    test_suite='reahl.component_dev',
    entry_points={
        'reahl.translations': [
            'reahl-component = reahl.messages'    ],
        'console_scripts': [
            'reahl-control = reahl.component.prodshell:ProductionCommandline.execute_one'    ],
        'reahl.eggs': [
            'Egg = reahl.component.eggs:ReahlEgg'    ],
        'reahl.component.prodcommands': [
            'CreateDBUser = reahl.component.prodshell:CreateDBUser',
            'DropDBUser = reahl.component.prodshell:DropDBUser',
            'CreateDB = reahl.component.prodshell:CreateDB',
            'DropDB = reahl.component.prodshell:DropDB',
            'BackupDB = reahl.component.prodshell:BackupDB',
            'RestoreDB = reahl.component.prodshell:RestoreDB',
            'BackupAllDB = reahl.component.prodshell:BackupAllDB',
            'RestoreAllDB = reahl.component.prodshell:RestoreAllDB',
            'SizeDB = reahl.component.prodshell:SizeDB',
            'RunJobs = reahl.component.prodshell:RunJobs',
            'CreateDBTables = reahl.component.prodshell:CreateDBTables',
            'DropDBTables = reahl.component.prodshell:DropDBTables',
            'MigrateDB = reahl.component.prodshell:MigrateDB',
            'DiffDB = reahl.component.prodshell:DiffDB',
            'ListConfig = reahl.component.prodshell:ListConfig',
            'CheckConfig = reahl.component.prodshell:CheckConfig',
            'ListDependencies = reahl.component.prodshell:ListDependencies'    ],
        'reahl.component.databasecontrols': [
            'NullDatabaseControl = reahl.component.dbutils:NullDatabaseControl'    ],
                 },
    extras_require={},
    cmdclass={'install_test_dependencies': InstallTestDependencies}
)
